/*
Copyright (C) 2014, 2015, 2016 Paul Boddie

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
*/

module cartridge()
{
	$fn = 50;

	/* Model configurations. Do not edit! */

	ROM_CARTRIDGE = 0;
	WIDE_CARTRIDGE = 1;

	/* Set this to the desired model type. */

	MODEL = ROM_CARTRIDGE;

	/*
	Configure the generated shapes.
	For the absolute simplest shape (for basic 3D printers)...

	BACK_CONNECTOR_SECTION = 0;
	FRONT_LABEL_INSET = 0; TOP_LABEL_INSET = 0;
	GROOVE = 0; ROUND_EDGES = 0; ROUND_CONNECTING_EDGES = 0;

	To retain the back connecting piece instead, change the above to...

	BACK_CONNECTOR_SECTION = 1;
	EXTERNAL_BACK_CAVITY = 0;
	EXTERNAL_BACK_CUTOUT = 1;

	Leave all options enabled for more capable 3D printers.
	*/

	/* The back part of the connecting section of the cartridge. */
	BACK_CONNECTOR_SECTION = 1;

	/* A necessary exterior cavity to let the cartridge fit inside the slot. */
	EXTERNAL_BACK_CAVITY = 1;

	/* An alternative cutout to let the cartridge fit inside the slot. */
	EXTERNAL_BACK_CUTOUT = 0;

	/* Optional insets for labels. */
	FRONT_LABEL_INSET = 1;
	TOP_LABEL_INSET = 1;

	/* Optional groove between the pieces to match the Electron. */
	GROOVE = 1;

	/* Optional rounding of the external edges. */
	ROUND_EDGES = 1;

	/* Recommended rounding of connection edges. */
	ROUND_CONNECTING_EDGES = 1;

	/* Set this where a shorter PCB is to be used. */

	SHORT_PCB = 0;

	/*
	Options for checking. Some useful combinations...

	Check feature alignment: APART = 0; BACK_SURFACE = 0; FRONT_SURFACE = 0;
	Check PCB alignment:     PCB = 1; BACK_SURFACE = 0;
	*/

	BACK = 1; BACK_SURFACE = 1;
	FRONT = 1; FRONT_SURFACE = 1;
	TOP_SURFACE = 1;

	/* For printing: APART = 1; PCB = 0; */

	APART = 1;
	PCB = 0;

	/* To check overlaps: INTERSECT = 1; CLOSED = 1; */

	INTERSECT = 0;
	CLOSED = 0;

	/* Separation when not apart and not closed. */

	SEPARATION = 10;

	/* To save time (before printing): FILLET = 0; */

	FILLET = 1;

	/*
	Rounding/fillet radius and additional margin of subtracted
	material. The additional margin helps avoid geometry problems.
	*/

	rr = 2;
	ro = rr;
	extra = 0.1;

	groove_rr = 0.2;
	groove_ro = groove_rr;

	pcb_lug_rr = 0.5;
	pcb_lug_ro = pcb_lug_rr;

	/* A fillet performs rounding using a quarter segment of a cylinder. */

	module fillet(r, h) {
		if (FILLET)
			translate([0, 0, -h/2])
				difference() {
					cube([r + extra, r + extra, h + extra]);
					cylinder(r = r, h = h);
				}
	}

	/* A fillet justified using the axes. */

	module fillet_justified(r, h) {
		if (FILLET)
			difference() {
				cube([r + extra, r + extra, h + extra]);
				cylinder(r = r, h = h);
			}
	}

	/* A justified fillet with the extra material below the y-axis. */

	module fillet_partitioned(r, h) {
		if (FILLET)
			difference() {
				translate([0, 0, -extra])
					cube([r + extra, r + extra, h + extra]);
				cylinder(r = r, h = h);
			}
	}

	module fillet_torus(radius, rounding) {
		if (FILLET)
			difference() {
				cube_at((radius + extra) * 2, (radius + extra) * 2, rounding + extra,
					0, 0, 1,
					0, 0, 0);
				union() {
					rotate_extrude(convexity = 10)
						translate([radius - rounding, 0, 0])
							circle(r = rounding);
					cylinder(r = radius - rounding, h = rounding);
				}
			}
	}

	/*
	Justify an object of the given dimensions, according to the given
	factors (where 1 indicates moving the object to the positive side of an
	axis, and -1 indicates moving it to the negative side of an axis).

	NOTE: child should eventually be replaced by children.
	*/
	module justify(width, depth, height, wdir, ddir, hdir) {
		translate([
			wdir * width / 2,
			ddir * depth / 2,
			hdir * height / 2])
			child();
	}

	/*
	Make a cuboid of the given dimensions, justifying it according to the given
	factors, and moving it to the specified coordinates.

	NOTE: Usage of justify within this module will not work due to recursion
	NOTE: limitations in openscad, potentially removed in more recent
	NOTE: releases. Thus, the justify transform is merged in here.
	*/
	module cube_at(width, depth, height, wdir, ddir, hdir, x, y, z) {
		translate([
			x + wdir * width / 2,
			y + ddir * depth / 2,
			z + hdir * height / 2])
			cube([width, depth, height], center = true);
	}

	/* Model widths. */

	ROM_CARTRIDGE_int_payload_width = 86.0;
	WIDE_CARTRIDGE_int_payload_width = 140.0;

	/* Internal dimensions. */

	/* internal width in the payload section */
	int_payload_width =
		MODEL == ROM_CARTRIDGE ? ROM_CARTRIDGE_int_payload_width :
		MODEL == WIDE_CARTRIDGE ? WIDE_CARTRIDGE_int_payload_width :
			ROM_CARTRIDGE_int_payload_width;

	int_connector_width = 86.0;	/* limited to the Plus 1 socket dimensions */
	int_payload_depth = 14.0;	/* maximum depth in the payload section */
	int_connector_depth = 11.5;	/* maximum depth in the connector section */
	int_payload_height = 50.8;	/* space between the top and the floor */
	int_connector_height = 13.5; 	/* vertical offset of bottom/floor of payload area */

	/* Side thicknesses. */

	front = 2;
	payload_back = 1;		/* in the payload area the thickness is reduced */
	connector_back = 3.5;		/* the back cavity requires a thicker back wall */
	top = 3;
	side = 2; 			/* increased from 1.5 for 3D printing reliability */
	bottom = 1;			/* thickness of floor of payload area */

	/* How much extra depth the back provides for mating with the front. */

	groove_width_extra = 2.0;	/* the depth of the groove accommodating the front rim */
	front_back_overlap = 1.0;
	groove_width_overlap = front_back_overlap + groove_width_extra;

	/* Division of pieces into front and back, defining the extents. */

	int_front_depth = 4.5;
	front_depth = int_front_depth + front;
	int_payload_back_depth = int_payload_depth - int_front_depth + front_back_overlap;
	int_connector_back_depth = int_connector_depth - int_front_depth + front_back_overlap;
	back_depth = int_payload_back_depth + payload_back;

	/* Cartridge dimensions. */

	payload_width = int_payload_width + side + side;
	payload_height = top + int_payload_height;

	connector_width = int_connector_width + side + side;
	connector_height = bottom + int_connector_height;

	height = payload_height + bottom + int_connector_height;
	depth = int_payload_depth + front + payload_back;

	upper_extent = height / 2;
	lower_extent = -upper_extent;
	int_payload_upper_extent = upper_extent - top;
	int_payload_lower_extent = lower_extent + int_connector_height + bottom;

	/* Connector extents. */

	int_connector_right_extent = int_connector_width / 2;
	int_connector_left_extent = -int_connector_right_extent;
	connector_left_extent = int_connector_left_extent - side;
	connector_right_extent = int_connector_right_extent + side;

	/* Where the payload is wider than the connector, the payload expands to the left. */

	int_payload_right_extent = int_connector_right_extent;
	int_payload_left_extent = int_payload_right_extent - int_payload_width;
	payload_left_extent = int_payload_left_extent - side;
	payload_right_extent = int_payload_right_extent + side;
	payload_centre = (payload_left_extent + payload_right_extent) / 2;

	/* Cartridge surfaces. */

	front_side = side;
	front_left = front_side;
	front_right = front_side;
	back_side = side;
	back_left = back_side;
	back_right = back_side;

	/* Label details. */

	front_label_width = payload_width - side - side - 3.0;
	front_label_height = 46.0;
	front_label_depth = 1.0;
	front_label_offset_from_bottom = 19.5;
	front_label_centre = payload_centre;
	front_label_left_extent = front_label_centre - front_label_width / 2;

	top_label_width = front_label_width;
	top_label_height = 11.5; /* the height as seen from above */
	top_label_depth = front_label_depth;
	top_label_offset_from_front = 2.5;
	top_label_centre = payload_centre;
	top_label_left_extent = top_label_centre - front_label_width / 2;

	/*
	The groove around the sides and top. This is extended to allow the
	pieces to fit together, and this extension is generated regardless of
	whether the visible groove is enabled or not.
	*/

	groove_width_exposed = GROOVE ? 1.5 : 0;
	groove_width_normal = groove_width_exposed + front_back_overlap;
	groove_depth = 1.0; /* how deep the groove goes into each side */

	/* Additional cutting to mate the back and front. */

	top_groove_width = groove_width_overlap;
	top_groove_depth = 2.0;

	/*
	Space for the inner edge of the back inside the front.
	Offsets are measured from the outside surfaces.
	*/

	inner_top_front_cutout_width = int_payload_width;
	inner_top_front_cutout_depth = top_groove_width;
	inner_top_front_cutout_height = top - top_groove_depth;

	inner_payload_front_cutout_height = payload_height - top_groove_depth;
	inner_payload_front_cutout_width = front_side - groove_depth;
	inner_payload_front_cutout_depth = groove_width_overlap;

	inner_connector_front_cutout_height = connector_height;
	inner_connector_front_cutout_width = front_side - groove_depth;
	inner_connector_front_cutout_depth = groove_width_overlap;

	/*
	The back cavity is the indented part at the bottom of the back of the
	cartridge. It appears to be mechanically necessary, making sure that
	cartridges cannot be plugged in the wrong way round.
	*/

	back_cavity_width = 68.0;
	back_cavity_inner_width = 65.0;
	back_cavity_offset_from_inner_left = 9.0;
	back_cavity_inner_offset_from_inner_left = 10.5;
	back_cavity_height = 13.5;
	back_cavity_inner_height = 12.0;
	back_cavity_depth = 1.5;

	/*
	The effect of the cavity on the inside of the case. In principle, the
	interior of the case could be straight since the plastic guides of the
	Plus 1 socket are outside the case.
	*/

	inner_back_slope_depth = 2.5;
	inner_back_slope_width = inner_back_slope_depth;
	inner_back_slope_max_offset = 10.5;
	inner_back_slope_min_offset = inner_back_slope_max_offset - inner_back_slope_width;

	/* The tapering off of the inner back edge. */

	inner_back_edge_width = 69.0;
	inner_back_edge_height = 3.0;
	inner_back_edge_depth = 1.5;

	/* The tapering off of the inner front edge. */

	inner_front_edge_width = connector_width - front_side * 2;
	inner_front_edge_height = 3.0;
	inner_front_edge_depth = 1.5;

	/*
	The cutout in the floor of the back of the cartridge that accommodates
	the edge connector.
	*/

	edge_connector_cutout_back_depth = 3.0;
	edge_connector_cutout_back_width = 57.5;

	/*
	The cutouts in the floor of the front of the cartridge that produce a
	kind of tab that guides the edge connector into place in the back cutout.
	*/

	edge_connector_cutout_front_depth = front_back_overlap;
	edge_connector_cutout_front_width = (int_connector_width - edge_connector_cutout_back_width) / 2;

	/*
	Edge connectors are themselves 0.05" or approximately 1.27mm in
	thickness according to the Acorn Electron Cartridge Interface Specification
	(Acorn Support Application Group Note 014).
	*/

	/* Extra internal features. Depths include front and back surfaces. */

	pcb_back_support_width = 1.2;
	pcb_back_support_depth = back_depth -
		edge_connector_cutout_back_depth;
	pcb_back_support_height = height - int_connector_height - top - bottom;

	pcb_front_support_width = 1.2;
	pcb_front_support_depth = int_front_depth;
	pcb_front_support_height = pcb_back_support_height;

	/*
	Features measured from the Stardot Dual ROM Adaptor cartridge board
	dimensions diagram. Offsets are from inside the bottom "floor".
	*/

	pcb_back_support_bump_width = pcb_front_support_width;
	pcb_back_support_bump_depth = 1.5;
	pcb_back_support_left_bump_height = 13.2;
	pcb_back_support_right_bump_height = 10.7;
	pcb_back_support_left_bump_offset_from_bottom = 15.1;
	pcb_back_support_right_bump_offset_from_bottom = 17.6;

	/* Configured by SHORT_PCB. */
	pcb_back_support_top_bump_height = 3.8;

	/* Move the PCB support towards the centre. */
	pcb_support_margin = 1.75;
	pcb_support_offset_from_centre = edge_connector_cutout_back_width / 2
		- pcb_support_margin;

	/*
	The PCB lugs protrude through the holes in the PCB. By extending both
	the front and back lugs by the depth of the back support bump, they
	overlap by that amount. The "mating" depth is defined as that amount
	plus an additional amount for adhesion.
	*/

	pcb_lug_mating_depth = pcb_back_support_bump_depth + 1.0;

	pcb_lug_depth = pcb_back_support_depth +
		pcb_lug_mating_depth;
	pcb_lug_inner_radius = 1.0;
	pcb_lug_outer_radius = 5.5 / 2;
	pcb_lug_offset_from_bottom = 14.35;
	pcb_lug_offset_from_inside = 5.55;

	/*
	The cross detail on the back lug is intended to resist the front lug,
	and thus starts at the point that the front lug ends.
	*/

	pcb_lug_cross_width = 6.7;
	pcb_lug_cross_depth = pcb_back_support_depth + pcb_back_support_bump_depth -
		pcb_lug_mating_depth;
	pcb_lug_cross_height = 1.4;

	pcb_front_lug_depth = pcb_front_support_depth +
		pcb_lug_mating_depth;
	pcb_front_lug_inner_radius = pcb_lug_outer_radius;
	pcb_front_lug_outer_radius = pow(
		pow(pcb_lug_cross_width / 2, 2) +
		pow(pcb_lug_cross_height / 2, 2),
		0.5);

	wide_pcb_lug_offset_from_inside = 6.0;
	wide_pcb_lug_offset_from_bottom = 6.0;

	/* PCBs for checking. */
 
	pcb_width = 84.85; pcb_height = 62.5; pcb_depth = 1;
	edge_connector_width = 56.5; edge_connector_height = 11.85;

	/*
	The rectangular part of the narrow left and right holes is smaller
	than the "bump" in the case, but the circular parts make the overall
	hole larger than the "bump". An extra depth is used for holes to avoid
	surface definition problems.
	*/

	pcb_hole_margin = 0.55;
	pcb_hole_width = 2.2;
	pcb_hole_extra_depth = 0.1;
	pcb_lug_hole_radius = 3.75;

	pcb_left_hole_offset = pcb_back_support_left_bump_offset_from_bottom + pcb_hole_margin;
	pcb_left_hole_height = pcb_back_support_left_bump_height - pcb_hole_margin * 2;
	pcb_left_hole_offset_from_centre = -(pcb_support_offset_from_centre +
		pcb_back_support_width / 2 - pcb_hole_width / 2);

	pcb_right_hole_offset = pcb_back_support_right_bump_offset_from_bottom + pcb_hole_margin;
	pcb_right_hole_height = pcb_back_support_right_bump_height - pcb_hole_margin * 2;
	pcb_right_hole_offset_from_centre = -pcb_left_hole_offset_from_centre;

	pcb_hole_depth = pcb_depth + 2 * pcb_hole_extra_depth;
	pcb_hole_start_depth = edge_connector_cutout_back_depth + pcb_hole_extra_depth;

	wide_pcb_width = 138.0; wide_pcb_height = pcb_height; wide_pcb_depth = pcb_depth;
	wide_pcb_hole_depth = pcb_hole_depth;
	wide_pcb_hole_start_depth = pcb_hole_start_depth;

	wide_pcb_lug_hole_radius = 4.0;

	/* An example feature for use with PCB testing. */

	feature_width = 15.0; feature_depth = 7.0; feature_height = 40.0;
	feature_height_offset = 10.0;

	/* Repeated constructs. */

	module pcb_support(xdir, bump_height, bump_offset) {

		/*
		Translate the support in the stated direction.
		Since the support is already justified in the direction, the translation
		moves the inner edge of the support to alignment with the end of the
		edge connector cutout and then back by the PCB support margin.
		*/

		translate([xdir *
			pcb_support_offset_from_centre,
			edge_connector_cutout_back_depth,
			int_payload_lower_extent])
			justify(pcb_back_support_width,
				pcb_back_support_depth,
				pcb_back_support_height,
				xdir, 1, 1)
				union() {

					/* Underlying support strut. */

					cube([pcb_back_support_width,
						pcb_back_support_depth,
						pcb_back_support_height], center = true);

					/* Middle bump. */

					cube_at(pcb_back_support_bump_width,
						pcb_back_support_bump_depth,
						bump_height,
						0, -1, 1,
						0,
						-pcb_back_support_depth / 2,
						-pcb_back_support_height / 2 + bump_offset);

					/* Top bump. */

					if (SHORT_PCB)
						cube_at(pcb_back_support_bump_width,
							pcb_back_support_bump_depth,
							pcb_back_support_top_bump_height,
							0, -1, 1,
							0,
							-pcb_back_support_depth / 2,
							pcb_back_support_height / 2 -
								pcb_back_support_top_bump_height);
				}
	}

	module pcb_lug(xdir) {
		pcb_lug_explicit(xdir, int_connector_width / 2 - pcb_lug_offset_from_inside,
			pcb_lug_offset_from_bottom);
	}

	module pcb_lug_wide(xdir) {
		pcb_lug_explicit(xdir, int_payload_width / 2 - wide_pcb_lug_offset_from_inside,
			wide_pcb_lug_offset_from_bottom);
		pcb_lug_explicit(xdir, int_payload_width / 2 - wide_pcb_lug_offset_from_inside,
			wide_pcb_height - edge_connector_height - wide_pcb_lug_offset_from_bottom);
	}

	module pcb_lug_explicit(xdir, pcb_lug_offset_from_centre, pcb_lug_offset_from_bottom) {
		translate([payload_centre + xdir * pcb_lug_offset_from_centre,
			back_depth,
			int_payload_lower_extent + pcb_lug_offset_from_bottom
			])
			rotate([90, 0, 0])
				difference() {

					/* Cylinder with cross. */

					union() {
						cylinder(h=pcb_lug_depth, r=pcb_lug_outer_radius);
						cube_at(pcb_lug_cross_width,
							pcb_lug_cross_height, pcb_lug_cross_depth,
							0, 0, 1,
							0, 0, 0);
						cube_at(pcb_lug_cross_height,
							pcb_lug_cross_width, pcb_lug_cross_depth,
							0, 0, 1,
							0, 0, 0);
					}

					/* Hollowed out by a cylinder. */

					cylinder(h=pcb_lug_depth, r=pcb_lug_inner_radius);

					/* Tapered off for easier connection. */

					if (ROUND_CONNECTING_EDGES)
						translate([0, 0, pcb_lug_depth - pcb_lug_ro])
							fillet_torus(pcb_lug_outer_radius, pcb_lug_rr);
				}
	}

	module pcb_front_lug(xdir) {
		pcb_front_lug_explicit(xdir, int_connector_width / 2 - pcb_lug_offset_from_inside,
			pcb_lug_offset_from_bottom);
	}

	module pcb_front_lug_wide(xdir) {
		pcb_front_lug_explicit(xdir, int_payload_width / 2 - wide_pcb_lug_offset_from_inside,
			wide_pcb_lug_offset_from_bottom);
		pcb_front_lug_explicit(xdir, int_payload_width / 2 - wide_pcb_lug_offset_from_inside,
			wide_pcb_height - edge_connector_height - wide_pcb_lug_offset_from_bottom);
	}

	module pcb_front_lug_explicit(xdir, pcb_lug_offset_from_centre, pcb_lug_offset_from_bottom) {
		translate([payload_centre + xdir * pcb_lug_offset_from_centre,
			-int_front_depth + pcb_front_lug_depth,
			int_payload_lower_extent + pcb_lug_offset_from_bottom
			])
			rotate([90, 0, 0])
				difference() {
					cylinder(h=pcb_front_lug_depth,
						r=pcb_front_lug_outer_radius);
					cylinder(h=pcb_front_lug_depth,
						r=pcb_front_lug_inner_radius);
				}
	}

	/* The actual shapes. */

	module front_piece(front_displacement, front_displacement_together) {
		if (FRONT)
			translate([front_displacement, front_displacement_together, 0])
			difference() {

				/* The cartridge surfaces. */

				union() {

					/* Front portion. */

					if (FRONT_SURFACE) {

						/* Surfaces surrounding the payload. */

						cube_at(payload_width, front, payload_height,
							0, -1, 1,
							payload_centre, -int_front_depth, int_payload_lower_extent);

						cube_at(front_left, front_depth, payload_height,
							-1, -1, 1,
							int_payload_left_extent, 0, int_payload_lower_extent);

						cube_at(front_right, front_depth, payload_height,
							1, -1, 1,
							int_payload_right_extent, 0, int_payload_lower_extent);

						/* Surfaces surrounding the connector. */

						cube_at(connector_width, front, connector_height,
							0, -1, -1,
							0, -int_front_depth, int_payload_lower_extent);

						cube_at(front_left, front_depth, connector_height,
							-1, -1, -1,
							int_connector_left_extent, 0, int_payload_lower_extent);

						cube_at(front_right, front_depth, connector_height,
							1, -1, -1,
							int_connector_right_extent, 0, int_payload_lower_extent);

						/* Top surface for the front piece. */

						if (TOP_SURFACE)
							cube_at(payload_width, front_depth, top,
								0, -1, 1,
								payload_centre, 0, int_payload_upper_extent);
					}

					difference() {

						/* Floor of cartridge. */

						if (payload_width > connector_width) {
							cube_at(payload_width, front_depth, bottom,
								1, -1, 1,
								payload_left_extent, 0, lower_extent + int_connector_height);
						} else {
							cube_at(int_connector_width, int_front_depth, bottom,
								0, -1, 1,
								0, 0, lower_extent + int_connector_height);
						}

						/* Left cutout. */

						cube_at(edge_connector_cutout_front_width,
							edge_connector_cutout_front_depth,
							bottom,
							1, -1, 1,
							int_connector_left_extent,
							0,
							lower_extent + int_connector_height);

						/* Right cutout. */

						cube_at(edge_connector_cutout_front_width,
							edge_connector_cutout_front_depth,
							bottom,
							-1, -1, 1,
							int_connector_right_extent,
							0,
							lower_extent + int_connector_height);
					}

					/* PCB supports. */

					if (MODEL == ROM_CARTRIDGE) {
						cube_at(pcb_front_support_width,
								pcb_front_support_depth,
								pcb_front_support_height,
								-1, -1, 1,
								-pcb_support_offset_from_centre,
								0,
								int_payload_lower_extent);

						cube_at(pcb_front_support_width,
								pcb_front_support_depth,
								pcb_front_support_height,
								1, -1, 1,
								pcb_support_offset_from_centre,
								0,
								int_payload_lower_extent);

						/* Circular "lugs" to hold PCBs in place. */

						pcb_front_lug(-1);
						pcb_front_lug(1);
					}

					if (MODEL == WIDE_CARTRIDGE) {

						pcb_front_lug_wide(-1);
						pcb_front_lug_wide(1);
					}
				}

				/* Label insets. */

				union() {

					/* Front label. */

					if (FRONT_LABEL_INSET)
						translate([front_label_left_extent, -front_depth,
							lower_extent + front_label_offset_from_bottom])
							cube([front_label_width, front_label_depth,
								front_label_height]);

					/* Top label. See also the back piece. */

					if (TOP_LABEL_INSET)
						translate([top_label_left_extent,
							-front_depth + top_label_offset_from_front,
							upper_extent - top_label_depth])
							cube([top_label_width, top_label_height,
								top_label_depth]);
				}

				/* Inner front edge cavity. */

				translate([inner_front_edge_width / 2, -int_front_depth, lower_extent])
					rotate([0, -90, 0])
						linear_extrude(height = inner_front_edge_width)
							polygon([
								[-extra, -inner_front_edge_depth],
								[0, -inner_front_edge_depth],
								[inner_front_edge_height, 0],
								[-extra, 0],
								]);

				/* Inner grooves for the top and sides of the back portion. */

				cube_at(inner_top_front_cutout_width,
					inner_top_front_cutout_depth,
					inner_top_front_cutout_height,
					1, -1, 1,
					int_payload_left_extent, 0, int_payload_upper_extent);

				cube_at(inner_payload_front_cutout_width,
					inner_payload_front_cutout_depth,
					inner_payload_front_cutout_height,
					1, -1, 1,
					int_payload_right_extent, 0, int_payload_lower_extent);

				cube_at(inner_payload_front_cutout_width,
					inner_payload_front_cutout_depth,
					inner_payload_front_cutout_height,
					-1, -1, 1,
					int_payload_left_extent, 0, int_payload_lower_extent);

				/* Cutout to accept the back connector sides (or the floor of the back piece). */

				cube_at(inner_connector_front_cutout_width,
					BACK_CONNECTOR_SECTION ? inner_connector_front_cutout_depth : edge_connector_cutout_front_depth,
					BACK_CONNECTOR_SECTION ? inner_connector_front_cutout_height : bottom,
					1, -1, -1,
					int_connector_right_extent, 0, int_payload_lower_extent);

				cube_at(inner_connector_front_cutout_width,
					BACK_CONNECTOR_SECTION ? inner_connector_front_cutout_depth : edge_connector_cutout_front_depth,
					BACK_CONNECTOR_SECTION ? inner_connector_front_cutout_height : bottom,
					-1, -1, -1,
					int_connector_left_extent, 0, int_payload_lower_extent);

				/* Fillets to round off the edges. */

				union() {
					if (ROUND_EDGES) {

						/* Top left and right rounding. */
					
						translate([payload_left_extent + ro, -front_depth / 2, upper_extent - ro])
							rotate([0, 0, 180])
								rotate([90, 0, 0])
									fillet(rr, front_depth);
						translate([payload_right_extent - ro, -front_depth / 2, upper_extent - ro])
							rotate([90, 0, 0])
								fillet(rr, front_depth);

						/* Top front rounding. */
					
						translate([payload_centre, -front_depth + ro, upper_extent - ro])
							rotate([0, 0, 180])
								rotate([0, -90, 0])
									fillet(rr, payload_width);

						/* Edge rounding. */
					
						translate([payload_right_extent - ro, -front_depth + ro, int_payload_lower_extent - bottom - extra])
							rotate([0, 0, 270])
								fillet_justified(rr, payload_height + bottom + extra);

						translate([payload_left_extent + ro, -front_depth + ro, int_payload_lower_extent - bottom - extra])
							rotate([0, 0, 180])
								fillet_justified(rr, payload_height + bottom + extra);

						translate([connector_right_extent - ro, -front_depth + ro, lower_extent])
							rotate([0, 0, 270])
								fillet_partitioned(rr, connector_height - bottom);

						translate([connector_left_extent + ro, -front_depth + ro, lower_extent])
							rotate([0, 0, 180])
								fillet_partitioned(rr, connector_height - bottom);
					}
				}
			}
	}

	module back_piece(back_displacement, back_rotation) {
		if (BACK)
			translate([back_displacement, 0, 0])
			rotate([0, 0, back_rotation]) {
				difference() {

					/* The cartridge surfaces. */

					union() {

						/* Back portion. */

						if (BACK_SURFACE) {

							/* Surfaces surrounding the payload. */

							/* Payload section of back surface. */

							cube_at(payload_width, payload_back, payload_height,
								0, 1, 1,
								payload_centre, int_payload_back_depth, int_payload_lower_extent);

							cube_at(back_left, back_depth, payload_height,
								-1, 1, 1,
								int_payload_left_extent, 0, int_payload_lower_extent);

							cube_at(back_right, back_depth, payload_height,
								1, 1, 1,
								int_payload_right_extent, 0, int_payload_lower_extent);

							/* Surfaces surrounding the connector. */

							/* Connector section of back surface overlapping the floor. */

							if (BACK_CONNECTOR_SECTION) {
								cube_at(connector_width, connector_back, connector_height,
									0, 1, -1,
									0, int_connector_back_depth, int_payload_lower_extent);

								cube_at(back_left, back_depth, connector_height,
									-1, 1, -1,
									int_connector_left_extent, 0, int_payload_lower_extent);

								cube_at(back_right, back_depth, connector_height,
									1, 1, -1,
									int_connector_right_extent, 0, int_payload_lower_extent);
							}

							/* Top of back piece. */

							if (TOP_SURFACE)
								cube_at(payload_width, back_depth, top,
									0, 1, 1,
									payload_centre, 0, int_payload_upper_extent);

							/*
							The extension of the back to connect with the front. Here, the inside
							edges are rounded. The outside edges are rounded later.

							An extra overlapping measure is employed so that the filleting is
							continuous between the payload and connector portions. On the inside
							edges, the connector filleting is extended upwards.
							*/

							/* Left side of payload. */

							difference() {
								cube_at(back_left - groove_depth, groove_width_extra, inner_payload_front_cutout_height,
									-1, 1, 1,
									int_payload_left_extent, -groove_width_extra, int_payload_lower_extent);
								if (ROUND_CONNECTING_EDGES)
									translate([int_payload_left_extent - groove_ro, -groove_width_extra + groove_ro,
										int_payload_lower_extent])
										rotate([0, 0, -90])
											fillet_justified(groove_rr, inner_payload_front_cutout_height);
							}

							/* Right side of payload. */

							difference() {
								cube_at(back_right - groove_depth, groove_width_extra, inner_payload_front_cutout_height,
									1, 1, 1,
									int_payload_right_extent, -groove_width_extra, int_payload_lower_extent);
								if (ROUND_CONNECTING_EDGES)
									translate([int_payload_right_extent + groove_ro, -groove_width_extra + groove_ro,
										int_payload_lower_extent])
										rotate([0, 0, 180])
											fillet_justified(groove_rr, inner_payload_front_cutout_height);
							}

							if (BACK_CONNECTOR_SECTION) {

								/* Left side of connector (or floor extension). */

								difference() {
									cube_at(back_left - groove_depth, groove_width_extra,
										inner_connector_front_cutout_height,
										-1, 1, -1,
										int_connector_left_extent, -groove_width_extra, int_payload_lower_extent);
									if (ROUND_CONNECTING_EDGES)
										translate([int_connector_left_extent - groove_ro,
											-groove_width_extra + groove_ro,
											lower_extent])
											rotate([0, 0, -90])
												fillet_partitioned(groove_rr,
													inner_connector_front_cutout_height + extra);
								}

								/* Right side of connector (or floor extension). */

								difference() {
									cube_at(back_right - groove_depth, groove_width_extra,
										inner_connector_front_cutout_height,
										1, 1, -1,
										int_connector_right_extent, -groove_width_extra, int_payload_lower_extent);
									if (ROUND_CONNECTING_EDGES)
										translate([int_connector_right_extent + groove_ro,
											-groove_width_extra + groove_ro,
											lower_extent])
											rotate([0, 0, 180])
												fillet_partitioned(groove_rr,
													inner_connector_front_cutout_height + extra);
								}
							}

							/* Top side. */

							difference() {
								cube_at(payload_width - groove_depth * 2, groove_width_extra, top - top_groove_depth,
									0, 1, 1,
									payload_centre, -groove_width_extra, int_payload_upper_extent);
								if (ROUND_CONNECTING_EDGES)
									translate([payload_centre, -groove_width_extra + groove_ro,
										int_payload_upper_extent + groove_ro])
										rotate([0, 0, 180])
											rotate([0, 90, 0])
												fillet(groove_rr, payload_width - groove_depth * 2);
							}
						}

						difference() {

							/* Floor of cartridge. */

							if (payload_width > connector_width) {

								difference() {
									cube_at(payload_width, back_depth, bottom,
										1, 1, 1,
										payload_left_extent, 0, lower_extent + int_connector_height);

									/* Cut out the floor to provide a tab. */

									cube_at(payload_width - connector_width + groove_depth,
										edge_connector_cutout_front_depth, bottom,
										1, 1, 1,
										payload_left_extent, 0, lower_extent + int_connector_height);
								}

							} else {
								cube_at(connector_width, back_depth, bottom,
									0, 1, 1,
									0, 0, lower_extent + int_connector_height);
							}

							/* Edge connector cutout. */

							cube_at(edge_connector_cutout_back_width,
								edge_connector_cutout_back_depth,
								bottom,
								0, 1, 1,
								0, 0, lower_extent + int_connector_height);
						}

						/* PCB supports. */

						if (MODEL == ROM_CARTRIDGE) {
							pcb_support(-1, pcb_back_support_left_bump_height,
								pcb_back_support_left_bump_offset_from_bottom);
							pcb_support(1, pcb_back_support_right_bump_height,
								pcb_back_support_right_bump_offset_from_bottom);

							/* Circular "lugs" to hold PCBs in place. */

							pcb_lug(-1);
							pcb_lug(1);
						}

						if (MODEL == WIDE_CARTRIDGE) {

							pcb_lug_wide(1);
							pcb_lug_wide(-1);
						}
					}

					/* Label insets. */

					union() {

						/* Top label. See also the front piece. */

						if (TOP_LABEL_INSET)
							translate([top_label_left_extent,
								-front_depth + top_label_offset_from_front,
								upper_extent - top_label_depth])
								cube([top_label_width, top_label_height,
									top_label_depth]);
					}

					/* Top and side grooves, positioned in the back portion. */

					union() {

						/*
						Grooves are separated to permit a discontinuity with the
						wide cartridge and for control over the back section.
						*/

						cube_at(groove_depth, groove_width_normal, payload_height + bottom,
							1, 1, -1,
							payload_left_extent, 0, upper_extent);

						if (BACK_CONNECTOR_SECTION)
							cube_at(groove_depth, groove_width_normal, connector_height,
								1, 1, 1,
								connector_left_extent, 0, lower_extent);

						/* Right groove. */

						cube_at(groove_depth, groove_width_normal, height,
							-1, 1, 0,
							payload_right_extent, 0, 0);

						if (BACK_CONNECTOR_SECTION)
							cube_at(groove_depth, groove_width_normal, connector_height,
								-1, 1, 1,
								connector_right_extent, 0, lower_extent);

						/* Top grooves. */

						cube_at(payload_width, groove_width_normal, groove_depth,
							0, 1, -1,
							payload_centre, 0, upper_extent);

						cube_at(payload_width, top_groove_width, top_groove_depth,
							0, 1, -1,
							payload_centre, -groove_width_extra, upper_extent);
					}

					if (BACK_CONNECTOR_SECTION) {

						/* Provide an external cavity. */

						if (EXTERNAL_BACK_CAVITY)
							intersection() {

								/* From the bottom upwards. */

								translate([0, back_depth, lower_extent])
									linear_extrude(height = back_cavity_height)
										translate([int_connector_left_extent, 0, 0])
											polygon([
												[back_cavity_offset_from_inner_left, 0],
												[back_cavity_inner_offset_from_inner_left,
													-back_cavity_depth],
												[back_cavity_inner_offset_from_inner_left +
													back_cavity_inner_width,
													-back_cavity_depth],
												[back_cavity_offset_from_inner_left +
													back_cavity_width, 0]
												]);

								/* From left to right. */

								translate([back_cavity_width / 2, back_depth, lower_extent])
									rotate([0, -90, 0])
										linear_extrude(height = back_cavity_width)
											polygon([
												[-extra, -back_cavity_depth],
												[back_cavity_inner_height,
													-back_cavity_depth],
												[back_cavity_height, 0],
												[-extra, 0]
												]);
							}

						/* Or remove the back section entirely. */

						else if (EXTERNAL_BACK_CUTOUT)
							translate([0, back_depth, lower_extent])
								linear_extrude(height = back_cavity_height)
									translate([int_connector_left_extent, 0, 0])
										polygon([
											[back_cavity_offset_from_inner_left, 0],
											[inner_back_slope_min_offset,
												-connector_back + inner_back_slope_depth],
											[inner_back_slope_min_offset, -connector_back],
											[int_connector_width - inner_back_slope_min_offset,
												-connector_back],
											[int_connector_width - inner_back_slope_min_offset,
												-connector_back + inner_back_slope_depth],
											[back_cavity_offset_from_inner_left +
												back_cavity_width, 0]
											]);

						/* Inner back cavities making the back thinner at the sides. */

						translate([0, int_connector_back_depth, lower_extent])
							linear_extrude(height = int_connector_height)
								translate([int_connector_left_extent, 0, 0])
									polygon([
										[0, 0],
										[inner_back_slope_max_offset, 0],
										[inner_back_slope_min_offset,
											inner_back_slope_depth],
										[0, inner_back_slope_depth]
										]);

						translate([0, int_connector_back_depth, lower_extent])
							linear_extrude(height = int_connector_height)
								translate([int_connector_right_extent, 0, 0])
									polygon([
										[0, 0],
										[-inner_back_slope_max_offset, 0],
										[-inner_back_slope_min_offset,
											inner_back_slope_depth],
										[0, inner_back_slope_depth]
										]);

						/* Inner back edge cavity, tapering the bottom edge. */

						translate([inner_back_edge_width / 2,
							int_connector_back_depth + inner_back_edge_depth, lower_extent])
							rotate([0, -90, 0])
								linear_extrude(height = inner_back_edge_width)
									polygon([
										[-extra, -inner_back_edge_depth],
										[inner_back_edge_height, -inner_back_edge_depth],
										[0, 0],
										[-extra, 0]
										]);
					}

					/* Fillets to round off the edges. */

					union() {
						if (ROUND_EDGES) {

							/* Top left and right rounding. */
						
							translate([payload_left_extent + ro, back_depth / 2, upper_extent - ro])
								rotate([0, 0, 180])
									rotate([90, 0, 0])
										fillet(rr, back_depth);
							translate([payload_right_extent - ro, back_depth / 2, upper_extent - ro])
								rotate([90, 0, 0])
									fillet(rr, back_depth);
						
							/* Top back rounding. */
						
							translate([payload_centre, back_depth - ro, upper_extent - ro])
								rotate([0, -90, 0])
									fillet(rr, payload_width);

							/* Outer edge rounding. */
						
							translate([payload_right_extent - ro, back_depth - ro, int_payload_lower_extent - bottom - extra])
								fillet_justified(rr, payload_height + bottom + extra);

							translate([payload_left_extent + ro, back_depth - ro, int_payload_lower_extent - bottom - extra])
								rotate([0, 0, 90])
									fillet_justified(rr, payload_height + bottom + extra);

							if (BACK_CONNECTOR_SECTION) {
								translate([connector_right_extent - ro, back_depth - ro, lower_extent])
									fillet_partitioned(rr, connector_height - bottom);

								translate([connector_left_extent + ro, back_depth - ro, lower_extent])
									rotate([0, 0, 90])
										fillet_partitioned(rr, connector_height - bottom);
							}
						}

						if (ROUND_CONNECTING_EDGES) {

							/*
							Outer edge rounding of the back extension. This is done as a
							separate removal operation rather than occurring when creating the
							extension in order to ensure that the edges are actually rounded.

							An extra overlapping measure is employed so that the filleting is
							continuous between the payload and connector portions. On the outside
							edges, the payload filleting is extended downwards.
							*/

							translate([payload_left_extent + groove_depth + groove_ro, -groove_width_extra + groove_ro,
								int_payload_lower_extent - extra])
								rotate([0, 0, 180])
									fillet_justified(groove_rr, inner_payload_front_cutout_height + extra);

							translate([payload_right_extent - groove_depth - groove_ro, -groove_width_extra + groove_ro,
								int_payload_lower_extent - extra])
								rotate([0, 0, -90])
									fillet_justified(groove_rr, inner_payload_front_cutout_height + extra);

							/* Sides of connector. */

							translate([connector_left_extent + groove_depth + groove_ro, -groove_width_extra + groove_ro,
								lower_extent])
								rotate([0, 0, 180])
									fillet_partitioned(groove_rr, inner_connector_front_cutout_height);

							translate([connector_right_extent - groove_depth - groove_ro, -groove_width_extra + groove_ro,
								lower_extent])
								rotate([0, 0, -90])
									fillet_partitioned(groove_rr, inner_connector_front_cutout_height);

							/* Top of payload. */

							translate([payload_centre, -groove_width_extra + groove_ro,
								int_payload_upper_extent + inner_top_front_cutout_height - groove_ro])
								rotate([0, 0, 180])
									rotate([0, -90, 0])
										fillet(groove_rr, payload_width - groove_depth * 2);
						}
					}
				}

				/* PCBs for checking. */
			 
				if (PCB) if (MODEL == ROM_CARTRIDGE) {
					translate([0, 0, int_payload_lower_extent])
						difference() {

							/* Mock PCB. */

							union() {
								cube_at(pcb_width, pcb_depth,
									pcb_height - edge_connector_height,
									0, -1, 1,
									0, edge_connector_cutout_back_depth, 0);
								cube_at(edge_connector_width, pcb_depth,
									edge_connector_height,
									0, -1, -1,
									0, edge_connector_cutout_back_depth, 0);

								/* Feature. */

								cube_at(feature_width, feature_depth, feature_height,
									0, 1, 1,
									0, edge_connector_cutout_back_depth, feature_height_offset);
							}

							/* Holes for mounting. */

							union() {
								cube_at(pcb_hole_width, pcb_hole_depth, pcb_left_hole_height,
									-1, -1, 1,
									pcb_left_hole_offset_from_centre, pcb_hole_start_depth,
										pcb_left_hole_offset);
								translate([pcb_left_hole_offset_from_centre -
									pcb_hole_width / 2, pcb_hole_start_depth, pcb_left_hole_offset])
									rotate([90, 0, 0])
										cylinder(h=pcb_hole_depth, r=pcb_hole_width / 2);
								translate([pcb_left_hole_offset_from_centre -
									pcb_hole_width / 2, pcb_hole_start_depth,
									pcb_left_hole_offset + pcb_left_hole_height])
									rotate([90, 0, 0])
										cylinder(h=pcb_hole_depth, r=pcb_hole_width / 2);
							}

							union() {
								cube_at(pcb_hole_width, pcb_hole_depth, pcb_right_hole_height,
									1, -1, 1,
									pcb_right_hole_offset_from_centre, pcb_hole_start_depth,
										pcb_right_hole_offset);
								translate([pcb_right_hole_offset_from_centre +
									pcb_hole_width / 2, pcb_hole_start_depth, pcb_right_hole_offset])
									rotate([90, 0, 0])
										cylinder(h=pcb_hole_depth, r=pcb_hole_width / 2);
								translate([pcb_right_hole_offset_from_centre +
									pcb_hole_width / 2, pcb_hole_start_depth,
									pcb_right_hole_offset + pcb_right_hole_height])
									rotate([90, 0, 0])
										cylinder(h=pcb_hole_depth, r=pcb_hole_width / 2);
							}

							/* Holes for lugs. */

							translate([
								int_connector_left_extent + pcb_lug_offset_from_inside,
								pcb_hole_start_depth, pcb_lug_offset_from_bottom])
								rotate([90, 0, 0])
									cylinder(h=pcb_hole_depth, r=pcb_lug_hole_radius);

							translate([
								int_connector_right_extent - pcb_lug_offset_from_inside,
								pcb_hole_start_depth, pcb_lug_offset_from_bottom])
								rotate([90, 0, 0])
									cylinder(h=pcb_hole_depth, r=pcb_lug_hole_radius);
						}
				}

				if (PCB) if (MODEL == WIDE_CARTRIDGE) {
					translate([0, 0, int_payload_lower_extent])
						difference() {

							/* Mock PCB. */

							union() {
								cube_at(wide_pcb_width, wide_pcb_depth,
									wide_pcb_height - edge_connector_height,
									0, -1, 1,
									payload_centre, edge_connector_cutout_back_depth, 0);
								cube_at(edge_connector_width, wide_pcb_depth,
									edge_connector_height,
									0, -1, -1,
									0, edge_connector_cutout_back_depth, 0);

								/* Feature. */

								cube_at(feature_width, feature_depth, feature_height,
									0, 1, 1,
									0, edge_connector_cutout_back_depth, feature_height_offset);
							}

							/* Holes for lugs. */

							translate([
								payload_centre + int_payload_width / 2 - wide_pcb_lug_offset_from_inside ,
								wide_pcb_hole_start_depth,
								wide_pcb_height - edge_connector_height - wide_pcb_lug_offset_from_bottom])
								rotate([90, 0, 0])
									cylinder(h=wide_pcb_hole_depth, r=wide_pcb_lug_hole_radius);

							translate([
								payload_centre - int_payload_width / 2 + wide_pcb_lug_offset_from_inside ,
								wide_pcb_hole_start_depth,
								wide_pcb_height - edge_connector_height - wide_pcb_lug_offset_from_bottom])
								rotate([90, 0, 0])
									cylinder(h=wide_pcb_hole_depth, r=wide_pcb_lug_hole_radius);

							translate([
								payload_centre + int_payload_width / 2 - wide_pcb_lug_offset_from_inside ,
								wide_pcb_hole_start_depth,
								wide_pcb_lug_offset_from_bottom])
								rotate([90, 0, 0])
									cylinder(h=wide_pcb_hole_depth, r=wide_pcb_lug_hole_radius);

							translate([
								payload_centre - int_payload_width / 2 + wide_pcb_lug_offset_from_inside ,
								wide_pcb_hole_start_depth,
								wide_pcb_lug_offset_from_bottom])
								rotate([90, 0, 0])
									cylinder(h=wide_pcb_hole_depth, r=wide_pcb_lug_hole_radius);
						}
				}
			}
	}

	/*
	Place the back piece next to the front, with the internals facing out
	the same way, if APART is defined.
	*/

	front_displacement_together = CLOSED ? front_back_overlap : APART ? 0 : front_back_overlap - SEPARATION;
	front_displacement = APART ? -connector_width * 0.6 : 0;
	back_displacement = APART ? connector_width * 0.6 : 0;
	back_rotation = APART ? 180 : 0;

	if (INTERSECT) {
		intersection() {
			front_piece(0, front_displacement_together);
			back_piece(0, 0);
		}
	} else {
		front_piece(front_displacement, front_displacement_together);
		back_piece(back_displacement, back_rotation);
	}
}

cartridge();
